#include "Plugin.h"

#include <assert.h>

#define LOGSHADERNOTES

using namespace ffgl;
using namespace ffglex;

namespace ffglqs
{
Plugin::Plugin()
{
}

Plugin::~Plugin()
{
}

FFResult Plugin::InitGL( const ViewportStruct* viewPort )
{
	std::string fragmentShaderCode = CreateFragmentShader( fragmentShaderBase );
	if( !shader.Compile( vertexShaderCode, fragmentShaderCode ) )
	{
		DeinitGL();
		return FF_FAIL;
	}

	shader.Use();
	SetParamUniformLocations();

	if( !quad.Initialise() )
	{
		DeinitGL();
		return FF_FAIL;
	}

	if( Init() == FF_FAIL )
	{
		DeinitGL();
		return FF_FAIL;
	}

	//FFGL requires us to leave the context in a default state on return
	resetOpenGLState();

	//Use base-class init as success result so that it retains the viewport.
	return PluginBase::InitGL( viewPort );
}

FFResult Plugin::ProcessOpenGL( ProcessOpenGLStruct* processOpenGLData )
{
	FFResult result = FF_FAIL;

	UpdateAudioAndTime();

	if( shader.IsReady() )
	{
		shader.Use();
		SendDefaultParams( shader );
		SendParams( shader );
		Update();
		result = Render( processOpenGLData );
		resetOpenGLState();
	}

	ConsumeAllTriggers();

	return result;
}

FFResult Plugin::DeinitGL()
{
	shader.FreeGLResources();
	quad.Release();
	Clean();
	return FF_SUCCESS;
}

FFResult Plugin::Render( ProcessOpenGLStruct* inputTextures )
{
	shader.Use();
	quad.Draw();
	return FF_SUCCESS;
}

std::string Plugin::CreateFragmentShader( const std::string& base ) const
{
	std::string fragmentShaderCode = fragmentShaderCodeStart;

	int i = 0;
	while( i < params.size() )
	{
		const ParamBase* param         = params[ i ];
		const std::string& uniformName = param->GetUniformName();

		if( uniformName.empty() )
		{
			// No sense in adding a uniform if there is no name specified!
			i++;
			continue;
		}

		const ParameterType type = param->GetType();

		auto grouping   = ParamFloat::Grouping::NONE;
		auto paramFloat = dynamic_cast< const ParamFloat* >( param );
		if( paramFloat != nullptr )
			grouping = paramFloat->GetGrouping();

		/*
		if( IsRGBColor( i ) )
		{
			fragmentShaderCode += "uniform vec3 " + uniformName + ";\n";
			i += 3;
		}
		else if( IsHueColor( i ) )
		{
			fragmentShaderCode += "uniform vec4 " + uniformName + ";\n";
			i += 4;
		}
*/
		if( grouping != ParamFloat::Grouping::NONE )
		{
			FFUInt32 count = static_cast< FFUInt32 >( grouping );

			fragmentShaderCode += "uniform vec" + std::to_string( count ) + " " + uniformName + ";\n";
			i += count;
		}
		else if( type == ParameterType::FF_TYPE_BOOLEAN || type == ParameterType::FF_TYPE_EVENT )
		{
			fragmentShaderCode += "uniform bool " + uniformName + ";\n";
			i++;
		}
		else if( type == ParameterType::FF_TYPE_INTEGER )
		{
			fragmentShaderCode += "uniform int " + uniformName + ";\n";
			i++;
		}
		else if( type == ParameterType::FF_TYPE_OPTION )
		{
			// We could handle this in the final "else" case; making explicit case to match with SendParams()
			fragmentShaderCode += "uniform float " + uniformName + ";\n";
			i++;
		}
		else if( type == ParameterType::FF_TYPE_BUFFER )
		{
			auto paramBuff = dynamic_cast< const ParamBuffer* >( param );
			if( paramBuff )
			{
				const ParameterUsage usage = paramBuff->GetUsage();

				// Check usage
				if( usage == ParameterUsage::FF_USAGE_FFT )
				{
					// Passed as a vec4 (volume, bass, mid, high)
					fragmentShaderCode += "uniform vec4 " + uniformName + ";\n";
				}
				else
				{
					// TODO: ffgl@subpixels.com - ParameterUsage::FF_USAGE_STANDARD not handled
				}
			}

			i++;
		}
		else if( type == ParameterType::FF_TYPE_TEXT )
		{
			// Text parameters not passed to shader
			// Also shouldn't have uniformName set, but just in case it is, we still want to skip it
			i++;
		}
		else
		{
			// "Normal" values (floats, 0.0 to 1.0, including non-grouped colour parameters), float range values
			fragmentShaderCode += "uniform float " + uniformName + ";\n";
			i++;
		}
	}

	for( auto snippet : includedSnippets )
	{
		fragmentShaderCode += shader::snippets.find( snippet )->second;
	}

	fragmentShaderCode += base;
	return fragmentShaderCode;
}

void Plugin::SetParamUniformLocations()
{
	for( auto param : params )
	{
		const std::string& uniformName = param->GetUniformName();
		if( !uniformName.empty() )
		{
			const GLint location = shader.FindUniform( uniformName );

			if( location != -1 )
			{
				param->SetUniformLocation( location );
			}
			else
			{
#ifdef LOGSHADERNOTES
				printf( "Note: Parameter [%s]: no shader uniform [%s]\n", param->GetName().c_str(), uniformName.c_str() );
#endif
			}
		}
	}
}

void Plugin::UpdateAudioAndTime()
{
	// Update time and frame data
	frame++;
	auto t_now = std::chrono::high_resolution_clock::now();
	timeNow    = std::chrono::duration< float, std::milli >( t_now - t_start ).count() / 1000.0f;
	deltaTime  = timeNow - lastUpdate;
	lastUpdate = timeNow;

	// Update FFT data
	for( auto paramFFT : audioParams )
	{
		paramFFT->audio.Update( paramFFT->data );
	}
}

// 2019-07-12 ffgl@subpixels.com
// - handle ParameterType::FF_TYPE_BUFFER
// - handle ParameterType::FF_TYPE_OPTION
void Plugin::SendParams( Shader& shader ) const
{
	// Clamp to edge is broken in Resolume right now so disable it
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER );
	FFUInt32 i = 0U;
	while( i < params.size() )
	{
		const ParamBase* param      = params[ i ];
		const ParameterType type    = param->GetType();
		const GLint uniformLocation = param->GetUniformLocation();

		// Try to access as a ParamFloat (most parameter types)
		auto paramFloat = dynamic_cast< const ParamFloat* >( param );

		auto grouping = ( paramFloat != nullptr ) ? paramFloat->GetGrouping() : ParamFloat::Grouping::NONE;

		if( grouping != ParamFloat::Grouping::NONE )
		{
			// Skip processing the grouped parameters if there is no uniform to set
			if( uniformLocation == -1 )
			{
				i += static_cast< FFUInt32 >( grouping );
				continue;
			}

			// Number of params (including this one) remaining
			FFUInt32 remaining = static_cast< FFUInt32 >( params.size() ) - i;
			bool isHSBColor    = true;// Until proven otherwise

			// Use an HSBAColor struct for convenience; up to 4 parameters, may need to do HSBtoRGB conversion later
			HSBAColor hsba = { 0.0f, 0.0f, 0.0f, 0.0f };

			// Get the first value
			hsba.hue = paramFloat->GetFloatValue();
			if( paramFloat->GetType() != ParameterType::FF_TYPE_HUE )
				isHSBColor = false;

			// Unexpected condition (grouping specified for the last parameter)
			if( remaining == 1 )
			{
				shader.Set( uniformLocation, hsba.hue );
				i++;
				continue;
			}

			// Get the second value
			paramFloat = GetParam< ParamFloat >( i + 1 );
			hsba.sat   = paramFloat->GetFloatValue();
			if( paramFloat->GetType() != ParameterType::FF_TYPE_SATURATION )
				isHSBColor = false;

			// VEC2 grouping (or only 2 parameters remaining)
			if( grouping == ParamFloat::Grouping::VEC2 || remaining == 2 )
			{
				shader.Set( uniformLocation, hsba.hue, hsba.sat );
				i += 2;
				continue;
			}

			// Get the third value
			paramFloat = GetParam< ParamFloat >( i + 2 );
			hsba.bri   = paramFloat->GetFloatValue();
			if( paramFloat->GetType() != ParameterType::FF_TYPE_BRIGHTNESS )
				isHSBColor = false;

			// VEC3 grouping (or only 3 parameters remaining)
			if( grouping == ParamFloat::Grouping::VEC3 || remaining == 3 )
			{
				// Note: this case can be a set of RED, GREEN, BLUE parameters.
				// Using HSBAColor members is fine (they are just floats).

				if( isHSBColor && param->doAutoConversion )
				{
					// Do HSB to RGB conversion (no alpha value)
					RGBAColor rgb;
					HSVtoRGB( hsba.hue, hsba.sat, hsba.bri, rgb.r, rgb.g, rgb.b );
					shader.Set( uniformLocation, rgb.r, rgb.g, rgb.b );
				}
				else
				{
					shader.Set( uniformLocation, hsba.hue, hsba.sat, hsba.bri );
				}

				i += 3;
				continue;
			}

			// VEC4 grouping

			// Get the fourth value
			paramFloat = GetParam< ParamFloat >( i + 3 );
			hsba.alpha = paramFloat->GetFloatValue();
			if( paramFloat->GetType() != ParameterType::FF_TYPE_ALPHA )
				isHSBColor = false;

			if( isHSBColor && param->doAutoConversion )
			{
				// Do HSB to RGB conversion (copy alpha value)
				RGBAColor rgba;
				HSBtoRGB( hsba, rgba );
				shader.Set( uniformLocation, rgba.r, rgba.g, rgba.b, rgba.a );
			}
			else
			{
				shader.Set( uniformLocation, hsba.hue, hsba.sat, hsba.bri, hsba.alpha );
			}

			i += 4;
		}
		else if( type == ParameterType::FF_TYPE_BOOLEAN || type == ParameterType::FF_TYPE_EVENT )
		{
			if( uniformLocation != -1 )
			{
				auto paramBool = dynamic_cast< const ParamBool* >( param );
				if( paramBool )
					shader.Set( uniformLocation, static_cast< int >( paramBool->GetBoolValue() ) );
			}

			i++;
		}
		else if( type == ParameterType::FF_TYPE_INTEGER )
		{
			// Integers are defined with a range, but the value is not normalised (0.0 to 1.0)
			if( uniformLocation != -1 )
			{
				auto paramInteger = dynamic_cast< const ParamInteger* >( param );
				if( paramInteger )
					shader.Set( uniformLocation, paramInteger->GetIntegerValue() );
			}

			i++;
		}
		else if( type == ParameterType::FF_TYPE_OPTION )
		{
			if( uniformLocation != -1 )
			{
				auto paramOption = dynamic_cast< const ParamOption* >( param );
				if( paramOption )
					shader.Set( uniformLocation, paramOption->GetFloatValue() );
			}

			i++;
		}
		else if( type == ParameterType::FF_TYPE_BUFFER )
		{
			if( uniformLocation != -1 )
			{
				auto paramBuffer = dynamic_cast< const ParamBuffer* >( param );
				if( paramBuffer )
				{
					// Check usage
					if( paramBuffer->GetUsage() == ParameterUsage::FF_USAGE_FFT )
					{
						// Note: Buffer data (elements[]) is copied to fftData[] in UpdateAudioAndTime()
						const auto paramFFT = dynamic_cast< const ParamFFT* >( param );
						if( paramFFT )
						{
							const Audio& a = paramFFT->audio;
							// Passed as a vec4 (volume, bass, mid, high)
							shader.Set( uniformLocation, a.GetVolume(), a.GetBass(), a.GetMid(), a.GetHigh() );
						}
					}
					else
					{
						// TODO: ffgl@subpixels.com - ParameterUsage::FF_USAGE_STANDARD not handled
					}
				}
			}

			i++;
		}
		else if( type == ParameterType::FF_TYPE_TEXT )
		{
			// Text parameters not passed to shader
			i++;
		}
		else
		{
			// "Normal" values (floats, 0.0 to 1.0, including non-grouped colour parameters), float range values
			if( uniformLocation != -1 )
			{
				if( paramFloat )
					shader.Set( uniformLocation, paramFloat->GetFloatValue() );
			}

			i++;
		}
	}
}

void Plugin::SendDefaultParams( ffglex::Shader& shader ) const
{
	shader.Set( "time", timeNow );
	shader.Set( "deltaTime", deltaTime );
	shader.Set( "frame", frame );
	shader.Set( "resolution", (float)viewport.width, (float)viewport.height );
	shader.Set( "bpm", beatInfo.bpm );
	shader.Set( "phase", beatInfo.barPhase );
}

void Plugin::SetSampleRate( unsigned int _sampleRate )
{
	sampleRate = _sampleRate;
	for( auto paramFFT : audioParams )
	{
		paramFFT->audio.SetSampleRate( _sampleRate );
	}
}

void Plugin::SetFragmentShader( const std::string& base )
{
	fragmentShaderBase = base;
}

void Plugin::AddAudioParam( ParamFFT* param )
{
	PluginManager::AddParam( param );
	audioParams.push_back( param );
}

// ------------------------------------------------------------
// HSBA Color Params
// ------------------------------------------------------------

ParamsVec Plugin::AddHSBAColorParams( const std::string& name, const HSBAColor& defaultColor )
{
	ParamsVec hsba;
	AddHSBAColorParams( hsba, name, name, defaultColor );
	return hsba;
}

ParamsVec Plugin::AddHSBAColorParams( const std::string& name, const std::string& uniformName, const HSBAColor& defaultColor )
{
	ParamsVec hsba;
	AddHSBAColorParams( hsba, name, uniformName, defaultColor );
	return hsba;
}

void Plugin::AddHSBAColorParams( ParamsVec& hsba, const std::string& name, const HSBAColor& defaultColor )
{
	AddHSBAColorParams( hsba, name, name, defaultColor );
}

void Plugin::AddHSBAColorParams( ParamsVec& hsba, const std::string& name, const std::string& uniformName, const HSBAColor& defaultColor )
{
	hsba.hue   = new ParamFloat( name, ParameterType::FF_TYPE_HUE, uniformName, defaultColor.hue, ParamFloat::Grouping::VEC4 );
	hsba.sat   = new ParamFloat( name + " S", ParameterType::FF_TYPE_SATURATION, "", defaultColor.sat );
	hsba.bri   = new ParamFloat( name + " B", ParameterType::FF_TYPE_BRIGHTNESS, "", defaultColor.bri );
	hsba.alpha = new ParamFloat( name + " A", ParameterType::FF_TYPE_ALPHA, "", defaultColor.alpha );

	AddParam( hsba.hue );
	AddParam( hsba.sat );
	AddParam( hsba.bri );
	AddParam( hsba.alpha );
}

// ------------------------------------------------------------
// RGB Color Params (alpha always 1.0)
// ------------------------------------------------------------

ParamsVec Plugin::AddRGBColorParams( const std::string& name, const ffglex::RGBAColor& defaultColor )
{
	ParamsVec rgb;
	AddRGBColorParams( rgb, name, name, defaultColor );
	return rgb;
}

ParamsVec Plugin::AddRGBColorParams( const std::string& name, const std::string& uniformName, const ffglex::RGBAColor& defaultColor )
{
	ParamsVec rgb;
	AddRGBColorParams( rgb, name, uniformName, defaultColor );
	return rgb;
}

void Plugin::AddRGBColorParams( ParamsVec& rgb, const std::string& name, const ffglex::RGBAColor& defaultColor )
{
	AddRGBColorParams( rgb, name, name, defaultColor );
}

void Plugin::AddRGBColorParams( ParamsVec& rgb, const std::string& name, const std::string& uniformName, const ffglex::RGBAColor& defaultColor )
{
	rgb.red   = new ParamFloat( name + " R", ParameterType::FF_TYPE_RED, uniformName, defaultColor.r, ParamFloat::Grouping::VEC3 );
	rgb.green = new ParamFloat( name + " G", ParameterType::FF_TYPE_GREEN, "", defaultColor.g );
	rgb.blue  = new ParamFloat( name + " B", ParameterType::FF_TYPE_BLUE, "", defaultColor.b );
	rgb.alpha = nullptr;

	AddParam( rgb.red );
	AddParam( rgb.green );
	AddParam( rgb.blue );
}

// ------------------------------------------------------------
// RGBA Color Params
// ------------------------------------------------------------

ParamsVec Plugin::AddRGBAColorParams( const std::string& name, const ffglex::RGBAColor& defaultColor )
{
	ParamsVec rgba;
	AddRGBAColorParams( rgba, name, name, defaultColor );
	return rgba;
}

ParamsVec Plugin::AddRGBAColorParams( const std::string& name, const std::string& uniformName, const ffglex::RGBAColor& defaultColor )
{
	ParamsVec rgba;
	AddRGBAColorParams( rgba, name, uniformName, defaultColor );
	return rgba;
}

void Plugin::AddRGBAColorParams( ParamsVec& rgba, const std::string& name, const ffglex::RGBAColor& defaultColor )
{
	AddRGBAColorParams( rgba, name, name, defaultColor );
}

void Plugin::AddRGBAColorParams( ParamsVec& rgba, const std::string& name, const std::string& uniformName, const ffglex::RGBAColor& defaultColor )
{
	rgba.red   = new ParamFloat( name + " R", ParameterType::FF_TYPE_RED, uniformName, defaultColor.r, ParamFloat::Grouping::VEC3 );
	rgba.green = new ParamFloat( name + " G", ParameterType::FF_TYPE_GREEN, "", defaultColor.g );
	rgba.blue  = new ParamFloat( name + " B", ParameterType::FF_TYPE_BLUE, "", defaultColor.b );
	rgba.alpha = new ParamFloat( name + " A", ParameterType::FF_TYPE_ALPHA, "", defaultColor.a );

	AddParam( rgba.red );
	AddParam( rgba.green );
	AddParam( rgba.blue );
	AddParam( rgba.alpha );
}

// ------------------------------------------------------------
// RGB Range Params (alpha always 1.0)
// ------------------------------------------------------------

ParamsVec Plugin::AddRGBRangeParams( const std::string& name, const std::string& uniformName, float defaultValue, const RangeStruct range )
{
	ParamsVec rgb;
	AddRGBRangeParams( rgb, name, uniformName, defaultValue, range );

	return rgb;
}

void Plugin::AddRGBRangeParams( ParamsVec& rgb, const std::string& name, const std::string& uniformName, float defaultValue, const RangeStruct range )
{
	rgb.red   = new ParamRange( name + " R", ParameterType::FF_TYPE_STANDARD, uniformName, defaultValue, range, ParamFloat::Grouping::VEC3 );
	rgb.green = new ParamRange( name + " G", ParameterType::FF_TYPE_STANDARD, "", defaultValue, range );
	rgb.blue  = new ParamRange( name + " B", ParameterType::FF_TYPE_STANDARD, "", defaultValue, range );
	rgb.alpha = nullptr;

	AddParam( rgb.red );
	AddParam( rgb.green );
	AddParam( rgb.blue );
}

// ------------------------------------------------------------
// Detecting when params groups represent colours
// ------------------------------------------------------------

bool Plugin::IsHueColor( int index ) const
{
	bool enoughSpace = index + 3 < params.size();
	if( !enoughSpace )
		return false;

	bool isColorType =
		params[ index ]->GetType() == ParameterType::FF_TYPE_HUE &&
		params[ index + 1 ]->GetType() == ParameterType::FF_TYPE_SATURATION &&
		params[ index + 2 ]->GetType() == ParameterType::FF_TYPE_BRIGHTNESS &&
		params[ index + 3 ]->GetType() == ParameterType::FF_TYPE_ALPHA;

	return isColorType;
}

bool Plugin::IsRGBColor( int index ) const
{
	bool enoughSpace = index + 2 < params.size();
	if( !enoughSpace )
		return false;
	bool isColorType =
		params[ index ]->GetType() == ParameterType::FF_TYPE_RED &&
		params[ index + 1 ]->GetType() == ParameterType::FF_TYPE_GREEN &&
		params[ index + 2 ]->GetType() == ParameterType::FF_TYPE_BLUE;

	return isColorType;
}

// ------------------------------------------------------------
// Code snippet inclusion
// ------------------------------------------------------------

void Plugin::Include( shader::snippet_id snippet )
{
	if( includedSnippets.find( snippet ) != includedSnippets.end() )
		return;

	includedSnippets.insert( snippet );
	auto deps = shader::dependencies.find( snippet );
	if( deps == shader::dependencies.end() )
		return;
	for( auto dep : deps->second )
	{
		includedSnippets.insert( dep );
	}
}

void Plugin::Include( std::set< shader::snippet_id > snippets )
{
	for( auto snippet : snippets )
	{
		Include( snippet );
	}
}

// ------------------------------------------------------------
// Triggers
// ------------------------------------------------------------

void Plugin::ConsumeAllTriggers()
{
	for( auto param : params )
	{
		auto trigger = dynamic_cast< ParamTrigger* >( param );
		if( trigger )
		{
			trigger->Consume();
		}
	}
}

}// namespace ffglqs
